//
// $Id: seafood.cc,v 1.12 1999/11/29 15:06:16 voeckler Exp $
//
// Author:  Jens-S. Vckler <voeckler@rvs.uni-hannover.de>
//
// File:    newcm.hh
//          Thu Aug 05 1999
//
// (c) 1999 Lehrgebiet Rechnernetze und Verteilte Systeme
//          Universitt Hannover, Germany
//
// Permission to use, copy, modify, distribute, and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that (i) the above copyright notices and this permission
// notice appear in all copies of the software and related documentation,
// and (ii) the names of the Lehrgebiet Rechnernetze und Verteilte
// Systeme and the University of Hannover may not be used in any
// advertising or publicity relating to the software without the
// specific, prior written permission of Lehrgebiet Rechnernetze und
// Verteilte Systeme and the University of Hannover.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL THE LEHRGEBIET RECHNERNETZE UND VERTEILTE SYSTEME OR
// THE UNIVERSITY OF HANNOVER BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
// INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
// SOFTWARE.
//
// $Log: seafood.cc,v $
// Revision 1.12  1999/11/29 15:06:16  voeckler
// added options to create both forms of output, textual and db intermediate.
// more extensive logging for unknown hierarchy codes, added flag to warn
// about NONE entries which are counted as TCP entries.
//
// Revision 1.11  1999/10/29 14:13:19  voeckler
// added new -D parameter, moved output into separate modules and control
// functions, added -H parameter for decent names for output, fixed some
// endless loop bug in the whitespace gutting function.
//
// Revision 1.10  1999/09/02 10:22:19  voeckler
// the dns cache is now a pointer and not a reference any longer. Changes
// to the DNS and IRR helper interface due to the use of external helper
// processes. Some wording changes.
//
// Revision 1.9  1999/08/25 21:22:13  voeckler
// added distributions, added more comments, made some tables
// configurable, fixed bug with command line configuration file
// specification.
//
// Revision 1.8  1999/08/23 11:53:03  voeckler
// fixed bug in dns and irr server which resulted in looking up any
// negativly cached entry. Also added some more debug output to the
// cache code.
//
// Revision 1.7  1999/08/20 22:47:19  voeckler
// number of CPUs online is checked, and if two or more, and the
// input file is seekable, a filter input will be used for compressed
// files (speeds things). Now made the DNS cache a pointer, too.
//
// Revision 1.6  1999/08/18 11:12:52  voeckler
// extended gzip input and bzip2 input to use their respective low-level
// interface. Also added a filter input method for multi processor
// machines (currently experimental).
//
// Revision 1.5  1999/08/15 22:26:56  voeckler
// added prefill buffer possibility to PlainText class, so that
// any pre-read content must not be re-read.
//
// Revision 1.4  1999/08/08 09:07:42  voeckler
// added compile-time configurable libz and libbz2 support.
//
// Revision 1.3  1999/08/06 13:07:08  voeckler
// added stats to fsm01 in order to parse "CONNECT some.host:443" correctly.
//
// Revision 1.2  1999/08/05 22:09:30  voeckler
// made the config file location commandline configurable, added a force
// no lookups flag.
//
// Revision 1.1  1999/08/05 21:16:27  voeckler
// Initial revision
//
//
#if defined(__GNUC__) || defined(__GNUG__)
#pragma implementation
#endif

#include <stdio.h>
#include <errno.h>
#include <string.h>

#include <sys/types.h>
#include <sys/time.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <limits.h>
#include <sys/stat.h>
#include <locale.h> // LC_ALL, LC_CTYPE, LANG
#include <time.h> 

#include "typedefs.h"
#include "global.hh"
#include "counter.hh"
#include "ctype.hh"
#include "hash.hh"
#include "string.hh"
#include "input.hh"
#include "logline.hh"
#include "counters.hh"
#include "tools.hh"
#include "fsm01.hh"
#include "output.hh"
#include "dumpdb.hh"
#include "dns.hh"

static const char*  RCS_ID =
"$Id: seafood.cc,v 1.12 1999/11/29 15:06:16 voeckler Exp $";

// global var to communication with flex/bison
int parserErrorSeen;
extern int yyparse();

#if !defined(_SC_NPROCESSORS_ONLN) && defined(_SC_NPROC_ONLN)
#define _SC_NPROCESSORS_ONLN _SC_NPROC_ONLN
#endif

inline
unsigned 
cpus( void )
  // purpose: determine number of CPUs currently online
{
  long result = sysconf(_SC_NPROCESSORS_ONLN);
  return ( result == -1 ? 1 : result );
}

char*
knowsAbout( const char* what )
  // purpose: find the absolute path to a given program
{
  // obtain PATH information
  char* path = getenv("PATH");
  if ( path == 0 ) return 0;

  // obtain maximum size of a path information
  char local[4096];
  char* s = path;
  struct stat st;
  while ( s != 0 ) {
    
    char* t = strchr( s, ':' );
    if ( t == 0 ) t = path+strlen(path);

    // create path info
    size_t i = ( size_t(t-s) > sizeof(local)-1 ? sizeof(local)-1 : t-s );
    strncpy( local, s, i );
    local[i] = '\0';
    strncat( local, "/", sizeof(local) );
    strncat( local, what, sizeof(local) );

    if ( stat(local,&st) != -1 && 
	 ((geteuid() == st.st_uid && (st.st_mode & S_IXUSR) > 0 ) ||
	  (getegid() == st.st_gid && (st.st_mode & S_IXGRP) > 0 ) ||
	  (st.st_mode & S_IXOTH) > 0 ) ) return strdup(local);
    s = t+1;
  }

  return 0;
}

static
BaseInput*
obtainFile( const char* fn ) 
  // purpose: check the given file name for being of the 'right' kind
  //          and return either a gunzip or a plain input stream to it.
  // paramtr: fn (IN): file name
  // returns: opened input stream, of NULL in case of error
{
  // try to open the specified file
  int fd = (fn[0]=='-' && fn[1]=='\0') ?
    dup(STDIN_FILENO) :
    open( fn, O_RDONLY | O_NOCTTY );
  if ( fd == -1 ) {
    fprintf( stderr, "open(%s): %s\n", fn, strerror(errno) );
    return 0;
  }
  
  // check for gzipped input by reading a few byte (see: /etc/magic)
  unsigned char test[PIPE_BUF];
  int option;
  do {
    option = read( fd, test, PIPE_BUF );
  } while ( option == -1 && errno == EINTR );

  // see how much we have got, complain bitterly otherwise
  if ( option < 0 ) {
    perror( "read from input" );
    close(fd);
    return 0;
  } else if ( option < 4 ) {
    fputs( "read from input: premature end of input\n", stderr );
    close(fd);
    return 0;
  }

  static const char* unknownFormat = 
    "unknown input format (neither gzip, bzip2 nor plain)\n";
  BaseInput* result = 0;
  if ( ISDIGIT(test[0]) && ISDIGIT(test[1]) && ISDIGIT(test[2]) ) {
    // plain input, we do not need to rewind
    fputs( "# using internal plain input\n", stderr );
    result = new PlainInput( fd, 65520, (const char*) test, option );
  } else if ( cpus() > 1 && lseek( fd, -option, SEEK_CUR ) != -1 ) {
    // There are more than 1 CPU, so use an external program filter to
    // decompress the input. That way, we make use of more than 1 CPU.
    // Especially gzip input will be almost as fast as plain input.
    //    For filter input, we must reset the input stream, but that
    // will fail for pipe input or socket input.
    char* filter = 0;

    if ( (test[0] == 0x1F) && (test[1] == 0x8B) && (test[2] <= 0x08) &&
	 (filter=knowsAbout("gunzip")) != 0 ) {
      // gzip input
      fprintf( stderr, "# using external filter \"%s\"\n", filter );
      result = new FilterInput( fd, 65520, filter );
      free((void*) filter);
    } else if ( (test[0] == 0x42) && (test[1] == 0x5A) && (test[2] == 0x68) &&
		(test[3] >= 0x30) && (filter=knowsAbout("bunzip2")) != 0 ) {
      // bzip2 input
      fprintf( stderr, "# using external filter \"%s\"\n", filter );
      result = new FilterInput( fd, 65520, filter );
      free((void*) filter);
    } else if ( (test[0] == 0x1F) && (test[1] == 0x9D) && (test[2] >= 128) &&
		(filter=knowsAbout("uncompress")) != 0 ) {
      // compressed input
      fprintf( stderr, "# using external filter \"%s\"\n", filter );
      result = new FilterInput( fd, 65520, filter );
      free((void*) filter);
    } else {
      // unknown format
      fputs( unknownFormat, stderr );
    }

#ifdef USE_LIBZ
  } else if ( (test[0] == 0x1F) && (test[1] == 0x8B) && (test[2] <= 0x08) ) {
    // gzip input
    fputs( "# using internal zlib\n", stderr );
    result = new GZipInput( fd, 65520, (const char*) test, option );
#endif // USE_LIBZ
#ifdef USE_LIBBZ2
  } else if ( (test[0] == 0x42) && (test[1] == 0x5A) && (test[2] == 0x68) &&
	      (test[3] >= 0x30) ) {
    // bzip2 input
    fputs( "# using internal bzlib\n", stderr );
    result = new BZip2Input( fd, 65520, (const char*) test, option );
#endif // USE_LIBBZ2
  } else {
    // unknown input format
    fputs( unknownFormat, stderr );
  } // no such input

  return result;
}

double
processFile( BaseInput* in, Counters& cnt )
  // purpose: big loop around file for parsing the contents
  // paramtr: in (IO): input stream to read until it ends
  // returns: time spent processing this file
{
  char temp[8192];
  FSM_01 fsm;
  LogLine log;
  int urlsize, column;
  bool runOnce = true;
  String log_tld, log_sld, log_hierhost, log_client;

  MyUInt32 lastStamp = 0;	// copy of last timestamp
  unsigned long linenoCopy;	// file based line nr after LF was read

  char* const urlComponents[3] = { temp, temp+256, temp+(256+256) };
  const size_t urlCompSizes[3] = { 256, 256, 32 };

  // fixes for Squid false log messages
#ifdef REFRESH_FIX
  int tcp_refresh_hit = LogLine::stable.find("TCP_REFRESH_HIT") & 0xFFFF;
#endif // REFRESH_FIX
  int hier_none = LogLine::htable.find("NONE") | IS_NONE;

  double start = now();
  while ( ! in->eof() ) {
    // report the line we are currently looking at.
    if ( (in->lineno & 65535) == 0 )
      fprintf( stderr, "# %lu lines processed from this file\n", in->lineno );

    //
    // STAGE 1: start parsing a log line
    //
    log.shallow_clear();
    column = 0;

    // 1st column, timestamp, e.g. 931212014.216
    // just read the integer portion and discard fractional value
    if ( in->integer( log._stamp ) < 9 ) {
      fprintf( stderr, "%lu: illegal timestamp, skipping\n", in->lineno );
    } else {
      // we didn't read the '.xxx' part, so skip forward over it.
      in->skiptolws();
      // log._stamp += globals.gmtOffset; 
      column++;
    }

    // 2nd column, duration, e.g. 25812, or 0, or even -1
    if ( in->integer( log._duration ) == 0 ) {
      fprintf( stderr, "%lu: illegal duration, skipping\n", in->lineno );
    } else
      column++;


    // 3rd column, client address, e.g. "130.75.3.218" or "rot-lane.rvs...."
    if ( in->get( log._requester, sizeof(log._requester) ) < 3 ) {
      fprintf( stderr, "%lu: illegal client host address, skipping\n", 
	       in->lineno );
    } else
      column++;


    // 4th column, cache hit status, e.g.: "TCP_MISS/200"
    if ( log.parseStatus( in ) < 6 || log._statuscode == -1 ) {
      fprintf( stderr, "%lu: illegal return status, skipping\n", in->lineno );
    } else
      column++;


    // 5th column, reply size, e.g. 8838
    if ( in->integer( log._size ) == 0 ) {
      fprintf( stderr, "%lu: illegal reply size, skipping\n", in->lineno );
    } else
      column++;


    // 6th column, method string, e.g. "GET"
    if ( log.parseMethod( in ) < 3 || log._method == 0 ) {
      fprintf( stderr, "%lu: unknown method, skipping\n", in->lineno );
    } else
      column++;

    // 7th column, URI, e.g. "http://www.user.cityline.ru:8083/~itbshake/"
    if ( (urlsize = in->get( log._url, sizeof(log._url)) ) < 5 ) {
      fprintf( stderr, "%lu: strange URL, skipping\n", in->lineno );
    } else
      column++;
    
    // 8th column, ident, e.g. "-"
    if ( globals.noIdent ) {
      // precondition: no ident lookups, value will always be "-"
      // if it isn't, the URL contains whitespaces (which are illegal)
      // FIXME: what happens, if the line ends prematurely ?!
      int allow = 16;
      while ( in->get( temp, sizeof(temp) ) != 1 && *temp != '-' && 
	      allow > 0 && ! in->eof() ) {
	--allow;
#if 0 // NOT YET NEEDED, SAFE TIME HERE!
	// whoa! there were whitespaces in the url...
	strncat( log._url, "%20", sizeof(log._url) );
	strncat( log._url, temp, sizeof(log._url) );
	urlsize += strlen(temp) + 3;
#endif
	fprintf( stderr, "%lu: URL contains whitespaces, fixing!\n", 
		 in->lineno );
      }

      // too many whitespaces
      if ( allow == 0 ) {
	fprintf( stderr, "%lu: too many whitespaces in URL, skipping!\n",
		 in->lineno );
	column--;
      }

      // we found what we were looking for (hopefully).
      log._ident[0] = '-';
      log._ident[1] = '\0';
      column++;
    } else {
      // precondition: ident lookups --> url ws cognition system is down :-(
      if ( in->get( log._ident, sizeof(log._ident) ) == 0 ) {
	fprintf( stderr, "%lu: unknown ident string, skipping\n", in->lineno );
      } else
        column++;
    }

    // 9th column, hierarchy info, e.g. "DIRECT/www.user.cityline.ru"
    if ( log.parseHierarchy(in) == 0 || log._hierarchy == 0 ) {
      fprintf( stderr, "%lu: unknown hierarchy, skipping\n", in->lineno );
    } else
      column++;

    // 10th column, mime type, e.g. "text/html"
    // Illegal media types are 'valid' and counted as error.
    // Only parse media type for TCP et. al., skip for UDP
    if ( (log._status & IS_UDP) == 0 ) {
      if ( log.parseMediaType(in) == 0 ) {
	fprintf( stderr, "%lu: strange MIME type, skipping\n", in->lineno );
      } else
	column++;
    } else {
      column++;
    }

    // 11th and 12th column may contain request and reply headers, skip
    // also, skippasteoln() will update the line counter and positioner
    linenoCopy = in->lineno;
    in->skippasteoln();
    if ( column < 10 ) {
      log.dump( stderr, 10 );
      continue;
    }

    //
    // STAGE 2: convert some values into something even better computable
    //
    unsigned short log_port = 0;
    int log_scheme = 0;
    log_client = log._requester;

    if ( runOnce ) {
      if ( cnt.startTime == 0 || cnt.finalTime > log._stamp )
	cnt.startTime = log._stamp;
      runOnce = false;
    }
    if ( cnt.finalTime < log._stamp ) 
      cnt.finalTime = log._stamp;
    MyUInt32 log_peak = log._stamp / globals.peakInterval;

    if ( lastStamp && 
	 MyUInt32(abs(log._stamp-lastStamp)) > globals.warnCrashInterval ) {
      char oldstamp[32], newstamp[32];
      time_t t1 = lastStamp;
      time_t t2 = log._stamp;
      strftime( oldstamp, 32, "%d.%m.%Y %H:%M:%S %Z", localtime(&t1) );
      strftime( newstamp, 32, "%d.%m.%Y %H:%M:%S %Z", localtime(&t2) );
      fprintf( stderr, "%lu: detected %.0f s silence between two time stamps\n"
	       "+ last \"%s\", current \"%s\"\n", linenoCopy,
	       difftime(t2,t1), oldstamp, newstamp  );
    }
    lastStamp = log._stamp;

    if ( (log._status & IS_UDP) == 0 ) {
      // determine scheme, hostname (tld,2ld) and portnumber
      fsm.parse( log._url, urlComponents, urlCompSizes );
      if ( (log_scheme = LogLine::schemes.find(urlComponents[0])) == 0 &&
	   globals.warnUnknownScheme )
	fprintf( stderr, "%lu: unknown URL scheme \"%s\"\n", 
		 linenoCopy, urlComponents[0] );
      getDomainFromHost( urlComponents[1], log_tld, log_sld );
      if ( log_tld.length() == 0 || 
	   LogLine::domains.find(log_tld.c_str()) == 0 ) {
	if ( globals.warnUnknownTLD )
	  fprintf( stderr, "%lu: unknown TLD \"%s\" in hostname \"%s\"\n",
		   linenoCopy, log_tld.c_str(), urlComponents[1] );
	log_tld = str_error;
	log_sld = str_error;
      }

      log_port =
	ISDIGIT( urlComponents[2][0] ) ?
	strtol(urlComponents[2],0,10) & 0xFFFF :
	80; // FIXME: check with getservbyname()
      log_port = (( log_port == 80 ) ? 64 : ( log_port >> 10 ));
      log_hierhost = String( log._hierhost ).lower();
    }
    
    // use 'natural' sizes for indexing, should be fastest
    int log_status  = log._status & 0xFFFF;
    int log_method  = log._method & 0xFFFF;
    int log_media   = log._mediatype & 0xFFFF;
    int log_subtype = log._mediasubtype & 0xFFFF;

    // fix Squid false log messages
#define TCP_HIT (IS_HIT | IS_TCP)
#ifdef REFRESH_FIX
    if ( log_status == tcp_refresh_hit && log._statuscode == 200 )
      log._status &= ~IS_HIT;
    else
#endif // REFRESH_FIX
    if ( (log._status & TCP_HIT) == TCP_HIT &&
	 ( log._statuscode == 301 || log._statuscode == 302 ) )
      log._hierarchy = hier_none;

    //
    // STAGE 3: count!
    //
#ifdef LOOP_ADD
    cnt.all.incr( log._size, log._duration );
#endif
    cnt.methods[log_method ].incr( log._size, log._duration );

    if ( (log._status & IS_UDP) != 0 ) {
      //
      // UDP
      //
#ifdef LOOP_ADD
      cnt.udp.incr( log._size, log._duration );
#endif
      cnt.udp_client[ log_client ].incr( log._size, log._duration );
      cnt.peak_udp[ log_peak ].incr( log._size, log._duration );
      if ( (log._status & IS_HIT) != 0 ) {
	// count UDP HITs here
	cnt.udp_hit.incr( log._size, log._duration );
	cnt.udp_hit_client[ log_client ].incr( log._size, log._duration );
	cnt.udp_hit_status[ log_status ].incr( log._size, log._duration );
	cnt.hit_methods[ log_method ].incr( log._size, log._duration );
	cnt.peak_udp_hit[ log_peak ].incr( log._size, log._duration );
      } else {
	// count UDP MISSes here
	cnt.udp_miss.incr( log._size, log._duration );
	cnt.udp_miss_status[ log_status ].incr( log._size, log._duration );
      }

    } else {
      //
      // TCP and else
      //
      int log_hier = log._hierarchy & 0xFFFF;

      if ( (log._status & IS_TCP) == 0 && globals.warnNoneAsTCP ) {
	fprintf( stderr, "%lu: neither TCP nor UDP, counting as TCP\n", 
		 linenoCopy );
	log.dump(stderr,10);
      }

      // count all ***OTHER*** traffic here (not limited to TCP)
#ifdef LOOP_ADD
      cnt.tcp.incr( log._size, log._duration );
      cnt.tcp_client[ log_client ].incr( log._size, log._duration );
#endif
      cnt.tcp_tld[ log_tld ].incr( log._size, log._duration );
      cnt.tcp_sld[ log_sld ].incr( log._size, log._duration );
      cnt.tcp_scheme[ log_scheme ].incr( log._size, log._duration );
      cnt.tcp_mime[ log_media ].incr( log._size, log._duration );
      if ( cnt.nSubtype[log_media] )
	cnt.tcp_submime[ log_media ][ log_subtype ].incr( log._size, 
							  log._duration );
      cnt.tcp_ports[ log_port ].incr( log._size, log._duration );
      cnt.peak_tcp[ log_peak ].incr( log._size, log._duration );

      if ( (log._status & TCP_HIT) == TCP_HIT ) {
	//
	// TCP HITs
	//
	cnt.distribution[Counters::D_HIT].add( log._size, log._duration );

	cnt.tcp_hit.incr( log._size, log._duration );
	cnt.tcp_hit_client[ log_client ].incr( log._size, log._duration );
	cnt.tcp_hit_status[ log_status ].incr( log._size, log._duration );
	cnt.tcp_hit_tld[ log_tld ].incr( log._size, log._duration );
	cnt.tcp_hit_sld[ log_sld ].incr( log._size, log._duration );
	cnt.tcp_hit_mime[ log_media ].incr( log._size, log._duration );
	if ( cnt.nSubtype[log_media] )
	  cnt.tcp_hit_submime[ log_media ][ log_subtype ].incr( log._size, 
								log._duration );
	cnt.tcp_hit_scheme[ log_scheme ].incr( log._size, log._duration );
	cnt.hit_methods[ log_method ].incr( log._size, log._duration );

	cnt.tcp_hit_hier[ log_hier ].incr( log._size, log._duration );
	if ( (log._hierarchy & (IS_DIRECT|IS_NONE))==IS_DIRECT && log_hier )
	  cnt.tcp_hit_direct_host[ log_hierhost ].incr( log._size,
							log._duration );
	cnt.peak_tcp_hit[ log_peak ].incr( log._size, log._duration );
      } else if ( (log._hierarchy & IS_NONE) || log._hierarchy == 0 ) {
	//
	// MISSes with NONE (mostly errors)
	//
	cnt.distribution[Counters::D_NONE].add( log._size, log._duration );

	cnt.tcp_miss_none.incr( log._size, log._duration );
	cnt.tcp_miss_none_status[ log_status ].incr( log._size, log._duration );
	cnt.tcp_miss_none_client[ log_client ].incr( log._size, log._duration );
      } else {
	//
	// TCP MISSes
	//
	cnt.distribution[Counters::D_MISS].add( log._size, log._duration );

	cnt.tcp_miss.incr( log._size, log._duration );
	cnt.tcp_miss_status[ log_status ].incr( log._size, log._duration );
	cnt.tcp_miss_client[ log_client ].incr( log._size, log._duration );
      }

      //
      // for further TCP and alike traffic, look into the hierarchy
      //
      if ( log_hier && (log._hierarchy & IS_NONE) == 0 ) {
	// any kind of traffic with attached hierarchy code
#ifdef LOOP_ADD
	cnt.hier.incr( log._size, log._duration );
#endif
	//---cnt.hier_method[ log_hier ].incr( log._size, log._duration );
	// cnt.hier_host[ log_hierhost ].incr( log._size, log._duration );
	// cnt.hier_status[ log_hier ][ log_hierhost ].incr( log._size, 
	//						     log._duration );
	if ( (log._hierarchy & IS_DIRECT) != 0 ) {
	  // directly to destination transcentral
	  cnt.hier_direct.incr( log._size, log._duration );
	  cnt.hier_direct_method[ log_hier ].incr( log._size, log._duration );
	  cnt.hier_direct_host[ log_hierhost ].incr( log._size, log._duration );
	  cnt.peak_hier_direct[ log_peak ].incr( log._size, log._duration );
	} else if ( (log._hierarchy & IS_PEER) != 0 ) {
	  // any peer business (carp, cache digests, etc)
	  cnt.hier_peer.incr( log._size, log._duration );
	  cnt.hier_peer_method[ log_hier ].incr( log._size, log._duration );
	  cnt.hier_peer_host[ log_hierhost ].incr( log._size, log._duration );
	  cnt.hier_peer_status[ log_hier ][ log_hierhost ].incr( log._size, 
		log._duration );
	  cnt.peak_hier_peer[ log_peak ].incr( log._size, log._duration );
	} else if ( (log._hierarchy & IS_PARENT) != 0 ) {
	  cnt.hier_parent.incr( log._size, log._duration );
	  cnt.hier_parent_method[ log_hier ].incr( log._size, log._duration );
	  cnt.hier_peer_host[ log_hierhost ].incr( log._size, log._duration );
	  cnt.hier_peer_status[ log_hier ][ log_hierhost ].incr( log._size, 
		log._duration );
	  cnt.peak_hier_parent[ log_peak ].incr( log._size, log._duration );
	} else {
	  fprintf( stderr, 
		   "%lu: unknown hierarchy 0x%08x (not hier counted)!\n", 
		   linenoCopy, log._hierarchy );
	}
      }	// hierarchy traffic
    } // !UDP
  } // big loop

#ifndef LOOP_ADD
  // save cycles by *not* counting deducables values in the loop
  cnt.hier = cnt.hier_direct + cnt.hier_peer + cnt.hier_parent;
  cnt.tcp = cnt.tcp_hit + cnt.tcp_miss_none + cnt.tcp_miss;
  cnt.udp = cnt.udp_hit + cnt.udp_miss;
  cnt.all = cnt.udp + cnt.tcp;

  for ( CountMap::Iterator i(cnt.tcp_hit_client); i.avail(); i++ )
    cnt.tcp_client[i.key()] += i();
  for ( CountMap::Iterator i(cnt.tcp_miss_client); i.avail(); i++ )
    cnt.tcp_client[i.key()] += i();
  for ( CountMap::Iterator i(cnt.tcp_miss_none_client); i.avail(); i++ )
    cnt.tcp_client[i.key()] += i();
#endif

  fprintf( stderr, "# %lu lines processed from this file\n", in->lineno-1 );
  return now() - start;
}

void
readConfigFile( const char* fn )
{
  extern FILE* yyin;

  fprintf( stderr, "# trying to read \"%s\"\n", fn );
  if ( (yyin=fopen(fn,"r")) != 0 ) {
    yyparse();
    fclose(yyin);
    if ( parserErrorSeen ) {
      fprintf( stderr, "there were errors in the \"%s\" config file, exiting!\n",
	       fn );
      exit(1);
    }
  } else {
    fprintf( stderr, "unable to read config file \"%s\": %s, exiting!\n", 
	     fn, strerror(errno) );
    exit(1);
  }
  fprintf( stderr, "# done reading \"%s\"\n", fn );
}

static
void
helpMe( const char* argv0, const char* rcsid )
{
  // obtain my own basename...
  char* s = strrchr(argv0,'/');
  if ( s ) ++s;

  printf( "%s\nUsage:\t%s [-N] [-f cfgfile] [-D dbfile] [-O nr] [-H name] file1 [..]\n", rcsid, s );
  puts( " fileN\tyou need at least one input file, plain"
#ifdef USE_LIBZ
	"|gzip"
#endif
#ifdef USE_LIBBZ2
	"|bzip2"
#endif
	" format" );
  printf( " -f fn\tUse configuration file fn instead of %s.conf\n", argv0 );
  puts( " -N\tDisable all DNS and IRR lookups regardless of configuration "
	"content" );
  puts( " -D fn\tDump results into file fn for DB post-processing" );
  puts( " -O nr\tOutput option: none, table, dbase, both: see documentation");
  puts( " -H hn\tUse the cache hostname 'hn' in output, default: "
	"basename(file1)" );
  puts( "" );
  exit(1);
}

static
void
parseCommandline( int argc, char* argv[], 
		  char*& configFilename,
		  char*& databaseFile,
		  char*& outputCachename,
		  bool& forceDisableLookups,
		  MyUInt08& outputOption )
  // purpose: parse optional arguments from the command line
  // paramtr: argc (IN): see main()
  //          argv (IN): see main()
  //          configFilename (OUT): place of configuration file
  //          outputCachename (OUT): name of cache or cache group to use
  //          forceDisableLooups (OUT): true==disable DNS+IRR 
  //          outputOption (IO): 0=no output 1=human readable on stdout,
  //                             2=db output 3=(1)+(2)
{
  // check for any argument
  if ( argc <= 1 ) helpMe(argv[0],RCS_ID);

  // construct configuration file default name
  configFilename = new char[strlen(argv[0])+6];
  strcpy( configFilename, argv[0] );
  strcat( configFilename, ".conf" );
  forceDisableLookups = false;
  databaseFile = 0;

  // parse options
  struct stat st;
  int option;
  while ( (option = getopt( argc, argv, "D:f:H:NO:?" )) != -1 ) {
    switch ( option ) {
    case 'D':
      if ( databaseFile ) free((void*) databaseFile );
      databaseFile = 0;
      if ( optarg && *optarg ) {
	databaseFile = strdup(optarg);
	outputOption |= 0x02;
      }
      break;
    case 'f':
      if ( stat(optarg,&st) != -1 && 
	   ((geteuid() == st.st_uid && (st.st_mode & S_IRUSR) > 0 ) ||
	    (getegid() == st.st_gid && (st.st_mode & S_IRGRP) > 0 ) ||
	    (st.st_mode & S_IROTH) > 0 ) ) {
	delete[] configFilename;
	configFilename = new char[strlen(optarg)+1];
	strcpy( configFilename, optarg );
      } else {
	fprintf( stderr, "configuration file \"%s\" is invalid!\n", optarg );
	exit(1);
      }
      break;
    case 'H':
      if ( outputCachename ) free((void*) outputCachename );
      outputCachename = 0;
      if ( optarg && *optarg ) outputCachename = strdup(optarg);
      break;
    case 'N':
      forceDisableLookups = true;
      break;
    case 'O':
      outputOption = strtoul( optarg, 0, 0 ) & 0x03;
      break;
    case '?':
    default:
      helpMe(argv[0],RCS_ID);
      break;
    }
  }

  if ( outputCachename == 0 ) {
    // determine output cache name from first basename -- if there is one.
    // otherwise insist on having the output give a (virtual) name
    if ( optind == argc ) {
      fputs( "Sorry, but I need a (possibly virtual) name for the output\n"
	     "Please use the -H option to set such a name.\n", stderr );
      exit(1);
    }

    // determine the basename without suffix
    char* s = strrchr( argv[optind], '/' );
    if ( s == 0 ) s = argv[optind]; 
    else ++s;
    char* t = strchr( s, '.' );
    if ( t == 0 ) t = s + strlen(s);
#if 0
    fprintf( stderr, "# starting in \"%s\", going for %d chars.\n", s, t-s );
    exit(0);
#endif

    // copy the basename into outputCachename
    outputCachename = (char*) malloc( t-s+1 );
    memcpy( outputCachename, s, t-s );
    outputCachename[t-s] = '\0';
  }

  // ascertain that only legal host characters are used,
  // and replace illegal characters with an underscore.
  for ( char* s = outputCachename; s && *s; ++s )
    if ( ! ISHNAME(*s) ) *s = '_';
}

// -------------------------------------------------------------------

int
main( int argc, char* argv[] )
{
  double startup = now();
  double looptime = 0.0;
  MyUInt32 lineno = C_U32(0);

  // just a quick dash into locale settings
  if ( setlocale(LC_ALL,"C") == 0 ) 
    perror( "# unable to set C local environment" );

  // read some basic configuration commands
  bool disableLookups = false;
  char* configuration = 0;
  char* dumpFilename = 0;
  char* outputCachename = 0;
  MyUInt08 outputOption = 1;
  parseCommandline( argc, argv, configuration, dumpFilename, 
		    outputCachename, disableLookups, outputOption );
  readConfigFile( configuration );
  delete[] configuration;

  // force disable lookups, if such was commandline-wanted
  // that will still open the DNS database...
  if ( disableLookups ) {
    globals.clientFQDN = false;
    if ( globals.irrServer ) free((void*) globals.irrServer);
    globals.irrServer = 0;
  }

  // in case of debugging, use slower line buffered output
  if ( globals.debugClass || globals.debugLevel )
    setvbuf( stdout, 0, _IOLBF, 8192 );

  // *AFTER* reading configuration, allocate counters
  Counters counters; // default c'tor

  // try to read DNS cache 
  DNSCache* dns = 0;
  if ( ! disableLookups ) {
    fputs( "# trying to read DNS cache and start DNS co processes...\n",
	   stderr );
    dns = new DNSCache( globals.dnsChildren,
			globals.dnsHelper,
			globals.dnsCacheFile, 
			globals.dnsPositiveTTL, 
			globals.dnsNegativeTTL );
    fprintf( stderr, "# DNS cache contains %lu entries.\n", 
	     dns ? dns->size() : 0 );
  } else {
    fputs( "# no DNS cache required.\n", stderr );
  }

  // establish connection to IRR server
  IRRCache* irr = 0; // IRR connection
  if ( globals.irrServer && ! disableLookups ) {
    fprintf( stderr, "# read IRR cache, start IRR co processes, use server \"%s\".\n", 
	     globals.irrServer );
    irr = new IRRCache( globals.irrChildren,
			globals.irrHelper,
		        globals.irrCacheFile, 
			globals.irrServer,
			globals.irrPositiveTTL, 
			globals.irrNegativeTTL );
    fprintf( stderr, "# IRR cache contains %lu entries.\n", 
	     irr ? irr->size() : 0 );
  } else {
    fputs( "# no IRR services required.\n", stderr );
  }

  // process files
  for ( int filenr=optind; filenr < argc; filenr++ ) {
    BaseInput* in = obtainFile( argv[filenr] );
    if ( in ) { 
      looptime += processFile( in, counters );
      lineno += (in->lineno - 1);
    }
    delete in;
  }

  // show results
  fputs( "# done processing files, writing results\n", stderr );
  if ( (outputOption & 0x02) != 0 && dumpFilename ) {
    // module dumpdb.cc
    dumpResults( dumpFilename, outputCachename, 
		 dns, irr, counters, looptime, startup, lineno );
  }
  if ( outputOption & 0x01 ) {
    // module output.cc
    printResults( outputCachename, 
		  dns, irr, counters, looptime, startup, lineno );
  }

  // finalize local cache constructs
  if ( dns ) {
    fprintf( stderr, "# dumping DNS cache (%u entries)\n", dns->size() );
    delete dns;
    dns = 0;
  }

  if ( irr ) {
    fprintf( stderr, "# dumping IRR cache (%u entries)\n", irr->size() );
    delete irr;
    irr = 0;
  }

  fputs( "# DONE!\n", stderr );
  return 0;
}
